/*
* FunctionExpression.java - this is a built in function call.
*
* Copyright (c) 1996 Chuck McManis, All Rights Reserved.
*
* Permission to use, copy, modify, and distribute this software
* and its documentation for NON-COMMERCIAL purposes and without
* fee is hereby granted provided that this copyright notice
* appears in all copies.
*
* CHUCK MCMANIS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
* SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. CHUCK MCMANIS
* SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT
* OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
*/

package basic;

import java.util.Random;
import java.io.PrintStream;
import jp.gr.java_conf.sol.basic.*;

/**
* This class implements the mathematical functions for BASIC. The tokenizer
* will scan the input for one of the strings in <i>functions[]</i> and if
* it finds it, it returns a token of FUNCTION with its value being that of
* the index of the string into the array. (which are convientiently mapped
* into named constants).
*
* Parsing of the arguments to to the function are left up to the parse
* method in this class.
*/
class FunctionExpression extends Expression {

  Expression sVar; // STRING function variable.
  Random r;

  void print(PrintStream p) {
    p.print(SBasicCode.getFunction(oper));
    p.print("(");
    if (arg1 != null)
    arg1.print(p);
    if (arg2 != null)
    arg2.print(p);
  }

  public String toString() {
    return "FunctionExpression:: '"+unparse()+"'";
  }

  String unparse() {
    StringBuffer sb = new StringBuffer();

    sb.append(SBasicCode.getFunction(oper));
    sb.append("(");
    if (sVar != null) {
      sb.append(sVar.unparse());
      sb.append(", ");
    }
    if (arg1 != null) {
      sb.append(arg1.unparse());
      sb.append(", ");
    }
    sb.append(arg2.unparse());
    sb.append(")");
    return sb.toString();
  }

  private FunctionExpression(int t, Expression e) {
    super();
    oper = t;
    arg2 = e;
  }

  private FunctionExpression(int t, Expression a1, Expression a2) {
    super();
    arg1 = a1;
    arg2 = a2;
    oper = t;
  }

  double value(Program p) throws BASICRuntimeError {
    try {

      switch (oper) {
        case SBasicCode.FUNCTION_RAD :
        if (r == null)
        r = p.getRandom();
        return (r.nextDouble() * arg2.value(p));
        case SBasicCode.FUNCTION_INT :
        return Math.floor(arg2.value(p));
        case SBasicCode.FUNCTION_SIN :
        return Math.sin(arg2.value(p));
        case SBasicCode.FUNCTION_COS :
        return Math.cos(arg2.value(p));
        case SBasicCode.FUNCTION_TAN :
        return Math.tan(arg2.value(p));
        case SBasicCode.FUNCTION_ATN :
        return Math.atan(arg2.value(p));
        case SBasicCode.FUNCTION_SQR :
        return Math.sqrt(arg2.value(p));
        case SBasicCode.FUNCTION_ABS:
        return Math.abs(arg2.value(p));
        case SBasicCode.FUNCTION_LEN:
        BasicString s = arg2.stringValue(p);
        return (double) s.length();
        case SBasicCode.FUNCTION_LOG:
        return Math.log(arg2.value(p));
        case SBasicCode.FUNCTION_SGN:
        double v = arg2.value(p);
        if (v < 0) {
          return -1.0;
        } else if (v > 0) {
          return 1.0;
        }
        return 0.0;
        case SBasicCode.FUNCTION_VAL:
        Double dd = null;
        BasicString zz = (arg2.stringValue(p)).trim();
        try {
          dd = Double.valueOf(zz.getAsciiString());
        } catch (NumberFormatException nfe) {
          throw new BASICRuntimeError("Invalid string for VAL function.");
        }
        return dd.doubleValue();
        default :
        throw new BASICRuntimeError("Unknown or non-numeric function.");
      }
    } catch (Exception e) {
      if (e instanceof BASICRuntimeError)
      throw (BASICRuntimeError) e;
      else
      throw new BASICRuntimeError("Arithmetic Exception.");
    }
  }

  boolean isString() {
    switch (oper) {
      case SBasicCode.FUNCTION_LEFT$:
      case SBasicCode.FUNCTION_RIGHT$:
      case SBasicCode.FUNCTION_MID$:
      case SBasicCode.FUNCTION_CHR$:
      case SBasicCode.FUNCTION_STR$:
      return true;
      default:
      return false;
    }
  }

  BasicString stringValue(Program pgm) throws BASICRuntimeError {
    return stringValue(pgm, 0);
  }

  BasicString stringValue(Program pgm, int column) throws BASICRuntimeError {
    BasicString ss = null;
    int len = 0;

    if (sVar != null) {
      ss = sVar.stringValue(pgm);
      len = ss.length();
    }

    StringBuffer sb;
    int a;
    int code;

    switch (oper) {
      case SBasicCode.FUNCTION_LEFT$:
      code = (int)arg2.value(pgm);

      if (code >= 0) {
        if (code > ss.length()) {
          code = ss.length();
        }
        try {
          return ss.substring(0, code);
        } catch (Exception ex) {
          ex.printStackTrace();
        }
      } else {
        throw new BASICRuntimeError("ILLIGAL DATA ERROR");
      }
      case SBasicCode.FUNCTION_RIGHT$:
      code = (int)arg2.value(pgm);
      if (code >=0) {
        if (code > ss.length()) {
          code = ss.length();
        }
        try {
          return ss.substring(ss.length()-code, ss.length());
        } catch (Exception ex) {
          ex.printStackTrace();
        }
      } else {
        throw new BASICRuntimeError("ILLIGAL DATA ERROR");
      }

      return ss.substring(len - (int) arg2.value(pgm));
      case SBasicCode.FUNCTION_MID$:
      int t = (int) arg1.value(pgm);
      return ss.substring(t-1, (t-1)+(int) arg2.value(pgm));
      case SBasicCode.FUNCTION_CHR$:

      code = (int)arg2.value(pgm);
      if (code >= 0 && code <=255) {
        return new BasicString((char)code);
      } else {
        throw new BASICRuntimeError("ILLIGAL DATA ERROR");
      }

      case SBasicCode.FUNCTION_STR$:
      return new BasicString(arg2.value(pgm));
      default:
      return new BasicString("Function not implemented yet.");
    }

  }

  /**
  * Parse a function argument. This code pulls off the '(' and ')' around the
  * arguments passed to the function and parses them.
  */
  static FunctionExpression parse(int ty, LexicalTokenizer lt) throws BASICSyntaxError {
    FunctionExpression result;
    Expression a;
    Expression b;
    Expression se;
    Token t;

    t = lt.nextToken();
    if (! t.isSymbol('(')) {
      if (ty == SBasicCode.FUNCTION_RND) {
        lt.unGetToken();
        return new FunctionExpression(ty, new ConstantExpression(1));
      }
      throw new BASICSyntaxError("Missing argument for function.");
    }

    switch (ty) {
      case SBasicCode.FUNCTION_RND:
      case SBasicCode.FUNCTION_INT:
      case SBasicCode.FUNCTION_SIN:
      case SBasicCode.FUNCTION_COS:
      case SBasicCode.FUNCTION_TAN:
      case SBasicCode.FUNCTION_ATN:
      case SBasicCode.FUNCTION_SQR:
      case SBasicCode.FUNCTION_ABS:
      case SBasicCode.FUNCTION_CHR$:
      case SBasicCode.FUNCTION_VAL:
      case SBasicCode.FUNCTION_STR$:
      case SBasicCode.FUNCTION_LOG:
      a = ParseExpression.expression(lt);
      if (a instanceof BooleanExpression) {
        throw new BASICSyntaxError(SBasicCode.getFunction(ty)+" function cannot accept boolean expression.");
      }
      if ((ty == SBasicCode.FUNCTION_VAL) && (! a.isString()))
      throw new BASICSyntaxError(SBasicCode.getFunction(ty)+" requires a string valued argument.");
      result = new FunctionExpression(ty, a);
      break;

      case SBasicCode.FUNCTION_LEN:
      a = ParseExpression.expression(lt);
      if (! a.isString()) {
        throw new BASICSyntaxError(SBasicCode.getFunction(ty)+" function expects a string argumnet.");
      }
      result = new FunctionExpression(ty, a);
      break;
      case SBasicCode.FUNCTION_LEFT$:
      case SBasicCode.FUNCTION_RIGHT$:
      se = ParseExpression.expression(lt);
      if (! se.isString()) {
        throw new BASICSyntaxError(
        "Function expects a string expression.");
      }
      t = lt.nextToken();
      if (! t.isSymbol(',')) {
        throw new BASICSyntaxError(SBasicCode.getFunction(ty)+" function requires two arguments.");
      }
      a = ParseExpression.expression(lt);
      result = new FunctionExpression(ty, a);
      result.sVar = se;
      break;

      case SBasicCode.FUNCTION_MID$:
      se = ParseExpression.expression(lt);
      if (! se.isString()) {
        throw new BASICSyntaxError(
        "Function expects a string expression.");
      }
      t = lt.nextToken();
      if (! t.isSymbol(',')) {
        throw new BASICSyntaxError(SBasicCode.getFunction(ty)+ " function requires at least two arguments.");
      }
      a = ParseExpression.expression(lt);
      t = lt.nextToken();
      if (t.isSymbol(')')) {
        b = new ConstantExpression(1.0);
        lt.unGetToken();
      } else if (t.isSymbol(',')) {
        b = ParseExpression.expression(lt);
      } else {
        throw new BASICSyntaxError(SBasicCode.getFunction(ty)+ " unexpected symbol in expression.");
      }
      result = new FunctionExpression(ty, a, b);
      result.sVar = se;
      break;
      default:
      throw new BASICSyntaxError("Unknown function on input.");

    }
    t = lt.nextToken();
    if (! t.isSymbol(')')) {
      throw new BASICSyntaxError("Missing closing parenthesis for function.");
    }
    return result;
  }
}
