/*
* Program.java - One BASIC program, ready to roll.
*
* Copyright (c) 1996 Chuck McManis, All Rights Reserved.
*
* Permission to use, copy, modify, and distribute this software
* and its documentation for NON-COMMERCIAL purposes and without
* fee is hereby granted provided that this copyright notice
* appears in all copies.
*
* CHUCK MCMANIS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
* SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. CHUCK MCMANIS
* SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT
* OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
*/

package basic;
import java.io.*;
import java.util.Enumeration;
import java.util.Random;
import util.RedBlackTree;
import java.util.Stack;
import java.util.Vector;

import jp.gr.java_conf.sol.basic.*;
import jp.gr.java_conf.sol.basic.mz700.*;

/**
* This class instantiates a BASIC program. A valid program is one that is
* parsed and ready to run. You can run it by invoking the run() method.
* The standard input and output of the running basic program can either
* be passed into the <b>run</b> method, or they can be presumed to be the
* in and out streams referenced by the <b>System</b> class.
*
* This class uses Red-Black trees to hold the parsed statements and the
* symbol table.
*
* @author      Chuck McManis
* @version     1.1
* @see         CommandInterpreter
*
*/
public class Program {//implements Runnable {
  // this tree holds all of the statements.
  private RedBlackTree stmts = new RedBlackTree(new NumberCompare());

  // this tree holds all of the variables.
  private RedBlackTree vars = new RedBlackTree();

  private Stack stmtStack = new Stack();
  Vector dataStore = new Vector();
  int dataPtr = 0;
  Random r = new Random(0);
  String myName;

  boolean traceState = false;
  BasicPrintStream traceFile = null;
  private boolean shouldStop;
  private boolean running;
  private Statement breakPoint;

  public synchronized void setShouldStop(boolean newShouldStop) {
    shouldStop = newShouldStop;
  }

  public synchronized boolean isShouldStop() {
    return shouldStop;
  }

  void trace(boolean a) { traceState = a; }

  void trace(boolean a, BasicString f) {
    if (traceFile == null) {
      try {
        traceFile = new BasicPrintStream(new FileOutputStream(f.getAsciiString()), null);
      } catch (IOException e) {
        System.out.println("Couldn't open trace file.");
        traceFile = null;
      }
    }
    trace(a);
  }

  Random getRandom() {
    return r;
  }

  void randomize(double seed) {
    r = new Random((long) seed);
  }

  void randomize() {
    r = new Random(); // uses the clock
  }

  /** copy program*/
  public void setProgram(Program pgm2) {

    this.stmts = pgm2.stmts;
  }

  /**
  * There are two ways to create a new program object, you can load one from
  * an already open stream or you can pass in a file name and load one from
  * the file system.
  */
  public static Program load(InputStream source, BasicPrintStream out) throws IOException, BASICSyntaxError {
    BasicInputStream dis = null;
    dis = new BasicInputStream(new BufferedInputStream(source));
    char data[] = new char[256];
    LexicalTokenizer lt = new LexicalTokenizer(data);
    BasicString lineData;
    Statement s;
    Token t;
    Program result = new Program();

    while (true) {
      // read a line of our BASIC program.
      lineData = dis.readLine();

      // if EOF simply return.
      if (lineData.length() == 0)
      return result;

      // if the line was blank, ignore it.
      if (lineData.length() == 0)
      continue;

      lt.reset(lineData);
      t = lt.nextToken();
      if (t.typeNum() != Token.CONSTANT) {
        throw new BASICSyntaxError("Line failed to start with a line number.");
      }

      try {
        s = ParseStatement.statement(lt);
      } catch (BASICSyntaxError bse) {
        s = ParseStatement.parseAsUnknwonStatement(lt);
        System.out.println("CAN NOT PARSE:"+lineData);
        //bse.printStackTrace();
      }
      s.addText(lineData);
      s.addLine((int) t.numValue());
      result.add((int) t.numValue(), s);
      System.out.println(lineData);
    }
  }

  /**
  * Load the specified file and parse the basic statements it contains.
  * @throws IOException when the filename cannot be located or opened.
  * @throws BASICSyntaxError when the file does not contain a properly formed
  *          BASIC program.
  */
  public static Program load(BasicPrintStream out, BasicString fileName) throws IOException, BASICSyntaxError {
    // XXX this needs to use the SourceManager class //

    InputStream is = Tape.getTape4Load(fileName);
    if (is == null) {
      throw new IOException("FILE NOT FOUND.");
    }

    Program r = null;
    try {
      r = load(is, out);
    } catch (BASICSyntaxError e) {
      is.close();
      throw e;
    }
    is.close();
    return r;
  }

  /**
  * Write the basic program out to the passed output stream. Conceptually
  * this is identical to doing a list operation.
  */
  public void save(OutputStream out) throws IOException {
    BasicPrintStream p = new BasicPrintStream(out, null);
    list(p);
  }

  /**
  * Write the program out to the file named in <i>output</i>.
  */
  public void save(String output) throws IOException {
    FileOutputStream fos = new FileOutputStream(output);
    save(fos);
    fos.flush();
    fos.close();
  }

  /**
  * This method starts this program running in its own thread.
  */
  /*
  void start() {
    Thread t = new Thread(this);
    if (myName != null)
    t.setName(myName + " execution.");
    t.start();
  }*/

  /**
  * Add a statement to the current program. Statements are indexed
  * by line number. If the add fails for some reason this method
  * returns false.
  */
  boolean add(int line, Statement s) {
    Integer ln = new Integer(line);
    Object z = stmts.put(ln, s);
    return true;
  }

  /**
  * Delete a statement from the current program. Statements are
  * indexed by line numbers. If the statement specified didn't
  * exist, then this method returns false.
  */
  boolean del(int line) {
    if (stmts.remove(new Integer(line)) == null)
    return false;
    return true;
  }

  /** delete all program */
  void clear() {
    stmts = new RedBlackTree(new NumberCompare());
  }

  /** clear all variables*/
  void clearVars() {
    vars = new RedBlackTree();
  }

  /**
  * Compute the indices based on the expressions in the variable
  * object.
  */
  private int[] getIndices(Variable v) throws BASICRuntimeError {
    int result[] = new int[v.numExpn()];

    for (int i=0; i < result.length; i++) {
      result[i] = (int) v.expn(i).value(this);
    }
    return result;
  }

  /**
  * Return the numeric value of a variable in the symbol table.
  */
  double getVariable(Variable v) throws BASICRuntimeError {
    Variable vi = (Variable) vars.get(v.name);
    if (vi == null) {
      vi = v;
      vi.setValue(0);
      vars.put(vi.name, vi);
    }

    if (! vi.isArray())
    return vi.numValue();
    int ii[] = getIndices(v);
    return vi.numValue(ii);
  }

  /**
  * Return the contents of the string variable named <i>name</i>. If
  * the variable has not yet been declared (ie used) this method throws
  * a BASICRuntime error.
  */
  BasicString getString(Variable v) throws BASICRuntimeError {
    Variable vi = (Variable) vars.get(v.name);
    if (vi == null) {
      vi = v;
      vi.setValue(new BasicString(""));
      vars.put(vi.name, vi);
    }

    if (! v.isArray())
    return vi.stringValue();
    int ii[] = getIndices(v);
    return vi.stringValue(ii);
  }

  /**
  * Set the numeric variable <i>name</i> to have value <i>value</i>. If
  * this is the first time we have seen the variable, create a place for
  * it in the symbol table.
  */
  void setVariable(Variable v, double value) throws BASICRuntimeError {
    Variable vi = (Variable) vars.get(v.name);
    if (vi == null) {
      if (v.isArray())
      throw new BASICRuntimeError("Array must be declared in a DIM statement");
      vi = new Variable(v.name);
      vars.put(v.name, vi);
    }
    if (! vi.isArray()) {
      vi.setValue(value);
      return;
    }
    int ii[] = getIndices(v);
    vi.setValue(value, ii);
  }

  void setRandom(long seed) {
  }

  /**
  * Set the string variable named <i>name</i> to have the value <i>value</i>.
  * If this is the first use of the variable it is created.
  */
  void setVariable(Variable v, BasicString value) throws BASICRuntimeError {
    Variable vi = (Variable) vars.get(v.name);
    if (vi == null) {
      if (v.isArray())
      throw new BASICRuntimeError("Array must be declared in a DIM statement");
      vi = new Variable(v.name);
      vars.put(v.name, vi);
    }
    if (! vi.isArray()) {
      vi.setValue(value);
      return;
    }
    int ii[] = getIndices(v);
    vi.setValue(value, ii);
  }

  /**
  * This method is used by the DIM statement to DECLARE arrays. Given
  * the nature of arrays we force them to be declared before they can
  * be used. This is common to most BASIC implementations.
  */
  void declareArray(Variable v) throws BASICRuntimeError {
    Variable vi;
    int ii[] = getIndices(v);
    vi = new Variable(v.name, ii);
    Variable xx = (Variable) vars.put(v.name, vi);
  }

  /**
  * Compute and return the next program statement to be executed.
  * The policy is, if the current statement has another statement hanging
  * off its <i>nxt</i> pointer use that one, otherwise use the next one
  * in the program numerically.
  */
  Statement nextStatement(Statement s) {
    if (s == null) {
      return null;
    } else if (s.nxt != null) {
      return s.nxt;
    }
    return ((Statement) stmts.next(new Integer(s.line)));
  }


  /**
  * Return the statment whose line number is <i>line</i>
  */
  Statement getStatement(int line) {
    Statement s = (Statement) stmts.get(new Integer(line));
    return s;
  }


  /** return allStatement*/
  Enumeration getAllStatements() {
    return stmts.elements();
  }

  /** return first statement in program */
  Statement getFirstStatement() {
    Enumeration enu = stmts.elements();

    Statement s = null;
    if (enu.hasMoreElements()) {
      s = (Statement)enu.nextElement();
    }
    return s;
  }

  Statement getLastStatement() {
    Enumeration enu = stmts.elements();

    Statement s= null;
    while(enu.hasMoreElements()) {
      s = (Statement)enu.nextElement();
    }
    return s;
  }

  /**
  * List program lines from <i>start</i> to <i>end</i> out to the
  * PrintStream <i>p</i>. Note that due to a bug in the Windows
  * implementation of PrintStream this method is forced to append
  * a <CR> to the file.
  */
  void list(int start, int end, BasicPrintStream p) {

    for (Enumeration e = stmts.elements(); e.hasMoreElements(); ) {
      Statement s = (Statement) e.nextElement();
      if (!isShouldStop() && (s.lineNo() >= start) && (s.lineNo() <= end)) {
        p.write(s.asString());
        p.write("\n");
      }
    }
  }

  /**
  * This is the first variation on list, it simply list from the starting
  * line to the the end of the program.
  */
  void list(int start, BasicPrintStream p) {
    list(start, Integer.MAX_VALUE, p);
  }

  /**
  * This second variation on list will list the entire program to the passed
  * PrintStream object.
  */
  void list(BasicPrintStream p) {
    list(0, p);
  }

  /**
  * Run the program and use the passed in streams as its input and output streams.
  *
  * Prior to running the program the statement stack is cleared, and the data fifo
  * is also cleared. Thus re-running a stopped program will always work correctly.
  *
  * @throws BASICRuntimeError if an error occurs while running.
  */
  /*
  public void run(InputStream in, BasicPrintStream out) throws BASICRuntimeError {
    BasicPrintStream pout;
    Enumeration e = stmts.elements();
    stmtStack = new Stack();    // assume no stacked statements ...
    dataStore = new Vector();   // ...  and no data to be read.
    dataPtr = 0;
    Statement s;

    vars = new RedBlackTree();

    // if the program isn't yet valid.
    if (! e.hasMoreElements())
    return;

    if (out instanceof BasicPrintStream) {
      pout = (BasicPrintStream) out;
    } else {
      pout = new BasicPrintStream(out);
    }

    //
    while (e.hasMoreElements()) {
      s = (Statement) e.nextElement();
      if (s.keyword.equals("DATA")) {
        s.doit(this, in, pout);
      }
    }

    e = stmts.elements();
    s = (Statement) e.nextElement();
    do {
      int yyy;

      // While running we skip Data statements.
      try {
        yyy = in.available();
      } catch (IOException ez) {
        yyy = 0;
      }
      if (yyy != 0) {
        pout.write("Stopped at :"+s);
        pout.write("\n");
        push(s);
        break;
      }
      if (!s.keyword.equals("DATA")) {
        if (traceState) {
          s.trace(this, (traceFile != null) ? traceFile : pout);
        }

        s = s.doit(this, in, pout);
      } else
      s = nextStatement(s);
    } while (s != null);
  }
  */


  /**
  * This method resumes a program that has been stopped. If the program
  * wasn't really stopped it throws a BASICRuntimeError.
  *
  * @throws BASICRuntimeError - Program wasn't in a stopped state.
  */
  void resume(InputStream in, BasicPrintStream pout) throws BASICRuntimeError {
    Statement s;

    s = pop();
    if ((s == null) || (!s.keyword.equals("STOP"))){
      throw new BASICRuntimeError("RESUME ERROR");
    }
    s = nextStatement(s);
    do {
      s = s.doit(this, in, pout);
    } while (s != null);
  }

  void cont(InputStream in, BasicPrintStream pout) throws BASICRuntimeError {
    Statement s;
    int yyy;

    s = pop();
    do {
      /* While running we skip Data statements. */
      try {
        yyy = in.available();
      } catch (IOException e) {
        yyy = 0;
      }
      if (yyy != 0) {
        pout.write("Stopped at :"+s);
        pout.write("\n");
        push(s);
        break;
      }
      if (!s.keyword.equals("DATA"))  {
        if (traceState) {
          s.trace(this, (traceFile != null) ? traceFile : pout);
        }

        s = s.doit(this, in, pout);
      } else
      s = nextStatement(s);
    } while (s != null);
  }


  /*
  * These methods deal with pushing and popping statements from the statement
  * stack, and data items from the data stack.
  */

  /**
  * Push this statement on the stack (one of FOR, GOSUB, or STOP)
  */
  void push(Statement s) {
    stmtStack.push(s);
  }

  /**
  * Pop the next statement off the stack, return NULL if the stack is
  * empty.
  */
  Statement pop() {
    if (stmtStack.isEmpty())
    return null;
    return (Statement) stmtStack.pop();
  }

  /**
  * Add a token to the data FIFO.
  */
  void pushData(Token t) {
    dataStore.addElement(t);
  }

  /**
  * Get the next token in the FIFO, return null if the
  * FIFO is empty.
  */
  Token popData() {
    if (dataPtr > (dataStore.size() - 1))
    return null;
    return (Token) dataStore.elementAt(dataPtr++);
  }

  /**
  * Reset the data FIFO back to the beginning.
  */
  void resetData() {
    dataPtr = 0;
  }

  public synchronized void setRunning(boolean newRunning) {
    running = newRunning;
  }

  public synchronized boolean isRunning() {
    return running;
  }

  public void setBreakPoint(basic.Statement newBreakPoint) {
    breakPoint = newBreakPoint;
  }

  public basic.Statement getBreakPoint() {
    return breakPoint;
  }
}

